/**
   This program takes an Abstract Syntax Tree (AST)
   that is defined by the following grammar, and
   converts it into a string that contains a
   C/C++/Java like infix notation.

       Prog ::= Exp
              | '(' 'prog' (Fun | Exp)+ Exp ')'

        Fun ::= '(' 'fun' VARIABLE Lambda ')'   // a function declaration

     Lambda ::= '(' 'lambda' VARIABLE* Exp ')'  // formal parameters followed by function body

        Exp ::= Apply
              | If
              | While
              | Set
              | Var
              | Begin
              | Print
              | BExp
              | AExp
              | INTEGER
              | BOOLEAN
              | VARIABLE

      Apply ::= '(' 'apply' Exp Exp* ')'    // function value followed by actual parameters

         If ::= '(' 'if' Exp Exp Exp ')'

      While ::= '(' 'while' Exp Exp ')'

        Set ::= '(' 'set' VARIABLE Exp ')'

        Var ::= '(' 'var' VARIABLE Exp ')'

      Begin ::= '(' 'begin' Exp+ Exp ')'

      Print ::= '(' 'print' Exp ')'

       BExp ::= '(' '||'  Exp Exp+ ')'
              | '(' '&&'  Exp Exp+ ')'
              | '(' '!'   Exp ')'
              | '(' RelOp Exp Exp ')'
              | '('  EqOp Exp Exp ')'

      RelOp ::= '<' | '>' | '<=' | '>='
       EqOp ::= '==' | '!='

       AExp ::= '(' '+' Exp Exp* ')'
              | '(' '-' Exp Exp? ')'
              | '(' '*' Exp Exp+ ')'
              | '(' '/' Exp Exp  ')'
              | '(' '%' Exp Exp  ')'
              | '(' '^' Exp Exp  ')'

    INTEGER ::= [-][0-9]+
    BOOLEAN ::= 'true' | 'false'
   VARIABLE ::= [a-zA-Z][a-zA-Z0-9]*
*/

public class AST2infix
{
   public static String ast2infix(Tree tree)
   {
      return convertProg(tree, "");
   }


   // Convert a prog
   private static String convertProg(Tree tree, String indent)
   {
      String result = "";

      // Check whick kind of Prog we have.
      if ( ! tree.getElement().equals("prog") )
      {
         // Convert the single expression.
         result += convertExp(tree, indent);
      }
      else
      {
         result += "Program\n{\n";

         // Convert each expression.
         for (int i = 0; i < tree.degree(); i++)
         {
            result += indent+"   " + convertExp(tree.getSubTree(i), indent+"   ");

            if ( tree.getSubTree(i).getElement().equals("fun") )
            {  // give function definitions an extra \n
               result += "\n";
            }

            if ( tree.getSubTree(i).getElement().equals("fun")
              || tree.getSubTree(i).getElement().equals("begin")
              || tree.getSubTree(i).getElement().equals("if")
              || tree.getSubTree(i).getElement().equals("while")
              || tree.getSubTree(i).getElement().equals("for") )
            {  // the above expressions do not need a semicolon
               result += "\n";
            }
            else // include a semicolon
               result += ";\n";
         }

         result += indent + "}";
      }

      return result;
   }//convertProg()


   // Convert a function definition.
   private static String convertFun(Tree tree, String indent)
   {
      String result = "";

      // do the name of the function
      result += tree.getSubTree(0).getElement() + "(";
      // get the number of formal parameters from the lambda expression
      int numberOfParameters = tree.getSubTree(1).degree() - 1;
      // do the list of formal parameters from the lambda expression
      for (int i = 0; i < numberOfParameters; i++)
      {
         String variable = tree.getSubTree(1).getSubTree(i).getElement();
         result += variable;
         if (i < numberOfParameters - 1)
            result += ", ";
      }
      result += ")\n";
      // do the function body from the lambda expression
      if ( tree.getSubTree(1).getSubTree(numberOfParameters).getElement().equals("begin") )
      {
         result += indent + convertExp(tree.getSubTree(1).getSubTree(numberOfParameters), indent);
      }
      else
      {
         result += indent + "{\n";
         result += indent+"   ";
         result += convertExp(tree.getSubTree(1).getSubTree(numberOfParameters), indent+"   ");
         if ( tree.getSubTree(1).getSubTree(numberOfParameters).getElement().equals("if")
           || tree.getSubTree(1).getSubTree(numberOfParameters).getElement().equals("while")
           || tree.getSubTree(1).getSubTree(numberOfParameters).getElement().equals("for") )
         {  // the above expressions do not need a semicolon
            result += "\n" + indent + "}";
         }
         else  // include a semicolon
            result += ";\n" + indent + "}";
      }

      return result;
   }//convertFun()


   // Convert an expression
   private static String convertExp(Tree tree, String indent)
   {
      String result = "";

      String node = tree.getElement();

      if ( node.equals("apply") )
         result += convertApply(tree, indent);
      else if ( node.equals("fun") )
         result += convertFun(tree, indent);
      else if ( node.equals("if") )
         result += convertIf(tree, indent);
      else if ( node.equals("while") )
         result += convertWhile(tree, indent);
      else if ( node.equals("set") )
         result += convertSet(tree, indent);
      else if ( node.equals("var") )
         result += convertVar(tree, indent);
      else if ( node.equals("begin") )
         result += convertBegin(tree, indent);
      else if ( node.equals("print") )
         result += convertPrint(tree, indent);
      else if ( node.equals("&&") || node.equals("||")
             || node.equals("!")
             || node.equals("<")  || node.equals(">")
             || node.equals("<=") || node.equals(">=")
             || node.equals("==") || node.equals("!=")
             || node.equals("+")  || node.equals("-")
             || node.equals("*")  || node.equals("/")
             || node.equals("%")  || node.equals("^") )
         result += convertBRAexp(tree, indent);
      else if ( 0 == tree.degree() )
      {
         result += node;
      }
      else // shouldn't get here; just convert to an S-expression
         result += tree;

      return result;
   }//convertExp()


   // Convert a function call
   private static String convertApply(Tree tree, String indent)
   {
      String result = "";

       // do the function value
      result += convertExp(tree.getSubTree(0), indent) + "(";
      // do the actual parameters
      for (int i = 1; i < tree.degree(); i++)
      {
         result += convertExp(tree.getSubTree(i), indent+result.length());
         if (i < tree.degree() - 1)
            result += ", ";
      }
      result += ")";

      return result;
   }//convertApply()


   // Convert an if-expression
   private static String convertIf(Tree tree, String indent)
   {
      String result = "";

      result += "if (" + convertExp(tree.getSubTree(0), indent+"       ") + ")\n";

      if ( tree.getSubTree(1).getElement().equals("begin") )
      {
         result += indent + convertExp( tree.getSubTree(1), indent );
      }
      else if ( tree.getSubTree(1).getElement().equals("while")
             || tree.getSubTree(1).getElement().equals("if")
             || tree.getSubTree(1).getElement().equals("for") )
      {
         result += indent+"   " + convertExp(tree.getSubTree(1), indent+"   ");
      }
      else
      {
         result += indent+"   " + convertExp( tree.getSubTree(1), indent+"   ");
         result += ";";
      }

      result += "\n" + indent + "else\n";

      if ( tree.getSubTree(2).getElement().equals("begin") )
      {
         result += indent + convertExp( tree.getSubTree(2), indent);
      }
      else if ( tree.getSubTree(2).getElement().equals("while")
             || tree.getSubTree(2).getElement().equals("if")
             || tree.getSubTree(2).getElement().equals("for") )
      {
         result += indent+"   " + convertExp( tree.getSubTree(2), indent+"   ");
      }
      else
      {
         result += indent+"   " + convertExp( tree.getSubTree(2), indent+"   ") + ";";
      }

      return result;
   }//convertIf()


   // Convert a while-loop expression
   private static String convertWhile(Tree tree, String indent)
   {
      String result = "";

      result += "while (" + convertExp(tree.getSubTree(0), indent+"       ") + ")\n";

      if ( tree.getSubTree(1).getElement().equals("begin") )
      {
         result += indent + convertExp( tree.getSubTree(1), indent);
      }
      else if ( tree.getSubTree(1).getElement().equals("while")
             || tree.getSubTree(1).getElement().equals("if")
             || tree.getSubTree(1).getElement().equals("for") )
      {
         result += indent+"   " + convertExp( tree.getSubTree(1), indent+"   ");
      }
      else
      {
         result += indent+"   " + convertExp( tree.getSubTree(1), indent+"   ");
         result += ";";
      }

      return result;
   }//convertWhile()


   // Convert a set expression
   private static String convertSet(Tree tree, String indent)
   {
      String result = "";

      // convert the left-hand-side of the assignment expression
      result += convertExp(tree.getSubTree(0), indent);

      result += " = ";

      // convert the right-hand-side of the assignment expression
      result += convertExp(tree.getSubTree(1), indent + result.length());

      return result;
   }//convertSet()


   // Convert a var expression
   private static String convertVar(Tree tree, String indent)
   {
      String result = "";

      // convert the left-hand-side of the declaration expression
      result += "var " + convertExp(tree.getSubTree(0), indent);

      result += " = ";

      // convert the right-hand-side of the declaration expression
      result += convertExp(tree.getSubTree(1), indent + result.length());

      return result;
   }//convertVar()


   // Convert a begin expression
   private static String convertBegin(Tree tree, String indent)
   {
      String result = "";

      result += "{\n";

      // process each expression
      for(int i = 0; i < tree.degree(); i++)
      {
         result += indent+"   " + convertExp(tree.getSubTree(i), indent + "   ");
         if ( tree.getSubTree(i).getElement().equals("begin")
           || tree.getSubTree(i).getElement().equals("if")
           || tree.getSubTree(i).getElement().equals("while")
           || tree.getSubTree(i).getElement().equals("for") )
         {  // the above expressions do not need a semicolon
            result += "\n";
         }
         else
            result += ";\n";
      }

      result += indent + "}";

      return result;
   }//convertBegin()


   // Convert a print expression
   private static String convertPrint(Tree tree, String indent)
   {
      String result = "";

      result += "print(" + convertExp(tree.getSubTree(0), indent+"       ") + ")";

      return result;
   }//convertPrint()


   // Convert a boolean, relational, or arithmetic expression
   private static String convertBRAexp(Tree tree, String indent)
   {
      String result = "";

      if ( tree.degree() == 1 )  // unary operator
      {
         if ( tree.getSubTree(0).degree() >= 2 )
         {
            result += tree.getElement() + "(" + convertExp(tree.getSubTree(0), indent) + ")";
         }
         else
         {
            result += tree.getElement() + convertExp(tree.getSubTree(0), indent);
         }
      }
      else if ( tree.degree() == 2 )  // binary operator
      {
         int middlePrecedence = precedence( tree.getElement() );
         int leftPrecedence = precedence( tree.getSubTree(0).getElement() );

         if ( leftPrecedence > middlePrecedence )
         {
            result += "(" + convertExp(tree.getSubTree(0), indent) + ")";
         }
         else
         {
            result += convertExp(tree.getSubTree(0), indent);
         }

         result += " " + tree.getElement() + " ";

         int rightPrecedence = precedence( tree.getSubTree(1).getElement() );

         if ( rightPrecedence >= middlePrecedence )
         {
            result += "(" + convertExp(tree.getSubTree(1), indent) + ")";
         }
         else
         {
            result += convertExp(tree.getSubTree(1), indent);
         }
      }
      else  // operators with 3 or more operands
      {
         int operatorPrecedence = precedence( tree.getElement() );
         int leftPrecedence = precedence( tree.getSubTree(0).getElement() );

         if ( leftPrecedence > operatorPrecedence )
         {
            result += "(" + convertExp(tree.getSubTree(0), indent) + ")";
         }
         else
         {
            result += convertExp(tree.getSubTree(0), indent);
         }

         for (int i = 1; i < tree.degree(); i++)
         {
            result += " " + tree.getElement() + " ";

            int rightPrecedence = precedence( tree.getSubTree(i).getElement() );

            if ( rightPrecedence >= operatorPrecedence )
            {
               result += "(" + convertExp(tree.getSubTree(i), indent) + ")";
            }
            else
            {
               result += convertExp(tree.getSubTree(i), indent);
            }
         }
      }
      return result;
   }//convertBRAexp()



   // http://www.java-tips.org/java-se-tips/java.lang/what-is-java-operator-precedence.html
   // http://www.cs.princeton.edu/introcs/11precedence/
   private static int precedence(String op)
   {
      int result = 0;  // "highest" precedence (i.e., constants)

      if ( op.equals("++")
        || op.equals("--") )
      {
         result = 1;
      }
      else if ( op.equals("index") )
      {
         result = 1;
      }
      else if ( op.equals("apply") )
      {
         result = 1;
      }
      else if ( op.equals("rand") )
      {
         result = 1;
      }
      else if ( op.equals("print") )
      {
         result = 1;
      }
      else if ( op.equals("!")
             || op.equals("+++")
             || op.equals("---") )
      {
         result = 2;
      }
      else if ( op.equals("^") )
      {
         result = 3;
      }
      else if ( op.equals("*")
             || op.equals("/")
             || op.equals("%") )
      {
         result = 4;
      }
      else if ( op.equals("+")
             || op.equals("-") )
      {
         result = 5;
      }
      else if ( op.equals("<")
             || op.equals(">")
             || op.equals("<=")
             || op.equals(">=") )
      {
         result = 7;
      }
      else if ( op.equals("==")
             || op.equals("!=") )
      {
         result = 8;
      }
      else if ( op.equals("&&") )
      {
         result = 12;
      }
      else if ( op.equals("||") )
      {
         result = 13;
      }
      else if ( op.equals("set") )
      {
         result = 15;
      }
      else if ( op.equals("var") )
      {
         result = 15;
      }
      else if ( op.equals("for") )
      {
         result = 16;
      }
      else if ( op.equals("if") )
      {
         result = 16;
      }
      else if ( op.equals("while") )
      {
         result = 16;
      }
      else if ( op.equals("begin") )
      {
         result = 16;
      }

      return result;
   }//precedence()

}//AST2infix