/**
   This program parses a token stream for the following grammar
   and builds an Abstract Syntax Tree (AST).

       Prog ::= Exp
              | '(' 'prog' (Fun | Exp)+ Exp ')'

        Fun ::= '(' 'fun' VARIABLE Lambda ')'   // a function declaration

     Lambda ::= '(' 'lambda' VARIABLE* Exp ')'  // formal parameters followed by function body

        Exp ::= Apply
              | If
              | While
              | Set
              | Var
              | Begin
              | Print
              | BExp
              | AExp
              | INTEGER
              | BOOLEAN
              | VARIABLE

      Apply ::= '(' 'apply' Exp Exp* ')'    // function value followed by actual parameters

         If ::= '(' 'if' Exp Exp Exp ')'

      While ::= '(' 'while' Exp Exp ')'

        Set ::= '(' 'set' VARIABLE Exp ')'

        Var ::= '(' 'var' VARIABLE Exp ')'

      Begin ::= '(' 'begin' Exp+ Exp ')'

      Print ::= '(' 'print' Exp ')'

       BExp ::= '(' '||'  Exp Exp+ ')'
              | '(' '&&'  Exp Exp+ ')'
              | '(' '!'   Exp ')'
              | '(' RelOp Exp Exp ')'
              | '('  EqOp Exp Exp ')'

      RelOp ::= '<' | '>' | '<=' | '>='
       EqOp ::= '==' | '!='

       AExp ::= '(' '+' Exp Exp* ')'
              | '(' '-' Exp Exp? ')'
              | '(' '*' Exp Exp+ ')'
              | '(' '/' Exp Exp  ')'
              | '(' '%' Exp Exp  ')'
              | '(' '^' Exp Exp  ')'

    INTEGER ::= [-][0-9]+
    BOOLEAN ::= 'true' | 'false'
   VARIABLE ::= [a-zA-Z][a-zA-Z0-9]*
*/

public class Parser
{
   public static Tree parse(String expression) throws ParseException
   {
      Tokenizer tokens = new Tokenizer(expression);  // Tokenizer is defined
                                                     // in Tokenizer.java
      Tree result = getProg(tokens);  // parse the token stream

      if ( tokens.hasToken() )  // there shouldn't be any more tokens
         throw new ParseException("unexpected input: "+tokens.peekToken()+"\n"+tokens + "\n");

      return result;
   }//parse()


   // Parse a prog
   private static Tree getProg(Tokenizer tokens) throws ParseException
   {
      Tree result = null;

      // Check whick kind of Prog we have.
      if ( ! ( tokens.has2Token() && tokens.peek2Token().equals("prog") ) )
      {
         // Evaluate the single expression.
         result = getExp( tokens );
      }
      else
      {
         tokens.match("(");

         if ( ! tokens.match("prog") )
            throw new ParseException("expected 'prog': " + tokens);

         result = new Tree("prog");

         // Parse each Fun or Exp in the Prog.
         while ( tokens.hasToken() && (! tokens.peekToken().equals(")")) )
         {
            if ( tokens.has2Token() && tokens.peek2Token().equals("fun") )
            {
               result.addSubTree( getFun(tokens) );
            }
            else
            {
               result.addSubTree( getExp(tokens) );
            }
         }

         if ( (! tokens.hasToken()) || (! tokens.match(")")) )
            throw new ParseException("expected ')': " + tokens);
      }

      return result;
   }//getProg()


   // Parse a function definition.
   private static Tree getFun(Tokenizer tokens) throws ParseException
   {
      if ( ! tokens.match("(") )
         throw new ParseException("expected '(': " + tokens);

      if ( ! tokens.match("fun") )
         throw new ParseException("expected 'fun': " + tokens);

      Tree result = new Tree("fun");

      // parse the function name
      if ( tokens.hasToken() )
      {
         result.addSubTree( new Tree(tokens.nextToken()) );
      }
      else
      {
         throw new ParseException("expected function name: " + tokens);
      }

      // parse the lambda expression
      if ( tokens.hasToken() )
      {
         result.addSubTree( getLambda(tokens) );
      }
      else
      {
         throw new ParseException("expected lambda expression: " + tokens);
      }

      if ( (! tokens.hasToken()) || (! tokens.match(")")) )
         throw new ParseException("expected ')': " + tokens);

      return result;
   }//getFun()


   // Parse a lambda expression
   private static Tree getLambda(Tokenizer tokens) throws ParseException
   {
      if ( ! tokens.match("(") )
         throw new ParseException("expected '(': " + tokens);

      if ( ! tokens.match("lambda") )
         throw new ParseException("expected 'lambda': " + tokens);

      Tree result = new Tree("lambda");

      // this is not really correct since we are not distinguishing
      // between the formal parameters and the function body
      while ( tokens.hasToken() && (! tokens.peekToken().equals(")")) )
      {
         result.addSubTree( getExp(tokens) );
      }

      if ( (! tokens.hasToken()) || (! tokens.match(")")) )
         throw new ParseException("expected ')': " + tokens);

      return result;
   }//getFun()


   // Parse an expression
   private static Tree getExp(Tokenizer tokens) throws ParseException
   {
      Tree result = null;

      String tk = tokens.peekToken();

      if ( tk.equals("(") )
      {
         if ( tokens.has2Token() )
            tk = tokens.peek2Token();
         else
            throw new ParseException("expected an expression: " + tokens);

         if ( tk.equals("apply") )
            result = getApply(tokens );
         else if ( tk.equals("if") )
            result = getIf( tokens);
         else if ( tk.equals("while") )
            result = getWhile(tokens);
         else if ( tk.equals("set") )
            result = getSet(tokens);
         else if ( tk.equals("var") )
            result = getVar(tokens);
         else if ( tk.equals("begin") )
            result = getBegin(tokens);
         else if ( tk.equals("print") )
            result = getPrint(tokens);
         else if ( tk.equals("&&") || tk.equals("||")
                || tk.equals("!") )
            result = getBexp(tokens);  // boolean expression
         else if ( tk.equals("<")  || tk.equals(">")
                || tk.equals("<=") || tk.equals(">=")
                || tk.equals("==") || tk.equals("!=") )
            result = getRexp(tokens);  // relational operator
         else if ( tk.equals("+") || tk.equals("-")
                || tk.equals("*") || tk.equals("/")
                || tk.equals("%") || tk.equals("^") )
            result = getAexp(tokens);  // arithmetic expression
         else
            throw new ParseException("unexpected expression: " + tk);
      }
      else
         result = new Tree(tokens.nextToken());

      return result;
   }//getExp()


   // Parse a function application.
   private static Tree getApply(Tokenizer tokens) throws ParseException
   {
      if ( ! tokens.match("(") )
         throw new ParseException("expected '(': " + tokens);

      if ( ! tokens.match("apply") )
         throw new ParseException("expected 'apply': " + tokens);

      Tree result = new Tree("apply");

      while ( tokens.hasToken() && (! tokens.peekToken().equals(")")) )
      {
         result.addSubTree( getExp(tokens) );
      }

      if ( (! tokens.hasToken()) || (! tokens.match(")")) )
         throw new ParseException("expected ')': " + tokens);

      return result;
   }//getApply()


   // Parse an if-expression
   private static Tree getIf(Tokenizer tokens) throws ParseException
   {
      if ( ! tokens.match("(") )
         throw new ParseException("expected '(': " + tokens);

      if ( ! tokens.match("if") )
         throw new ParseException("expected 'if': " + tokens);

      Tree result = new Tree("if");

      if ( tokens.hasToken() )
         result.addSubTree( getExp(tokens) );
      else
         throw new ParseException("expected conditional expression: " + tokens);

      if ( tokens.hasToken() )
         result.addSubTree( getExp(tokens) );
      else
         throw new ParseException("expected then expression: " + tokens);

      if ( tokens.hasToken() )
         result.addSubTree( getExp(tokens) );
      else
         throw new ParseException("expected else expression: " + tokens);

      if ( ! ( tokens.hasToken() && tokens.match(")") ) )
         throw new ParseException("expected ')': " + tokens);

      return result;
   }//getIf()


   // Parse a while-loop expression
   private static Tree getWhile(Tokenizer tokens) throws ParseException
   {
      if ( ! tokens.match("(") )
         throw new ParseException("expected '(': " + tokens);

      if ( ! tokens.match("while") )
         throw new ParseException("expected 'while': " + tokens);

      Tree result = new Tree ("while");

      if ( tokens.hasToken() )
         result.addSubTree( getExp(tokens) );
      else
         throw new ParseException("expected conditional expression: " + tokens);

      if ( tokens.hasToken() )
         result.addSubTree( getExp(tokens) );
      else
         throw new ParseException("expected loop body: " + tokens);

      if ( (! tokens.hasToken()) || (! tokens.match(")")) )
         throw new ParseException("expected ')': " + tokens);

      return result;
   }//getWhile()


   // Parse a set expression
   private static Tree getSet(Tokenizer tokens) throws ParseException
   {
      if ( ! tokens.match("(") )
         throw new ParseException("expected '(': " + tokens);

      if ( ! tokens.match("set") )
         throw new ParseException("expected 'set': " + tokens);

      Tree result = new Tree("set");

      // parse the variable
      if ( tokens.hasToken() )
         result.addSubTree( new Tree(tokens.nextToken()) );
      else
         throw new ParseException("expected variable name: " + tokens);

      // parse the expression
      if ( tokens.hasToken() )
         result.addSubTree( getExp(tokens) );
      else
         throw new ParseException("expected variable value: " + tokens);

      if ( (! tokens.hasToken()) || (! tokens.match(")")) )
         throw new ParseException("expected ')': " + tokens);

      return result;
   }//getSet()


   // Parse a var expression
   private static Tree getVar(Tokenizer tokens) throws ParseException
   {
      if ( ! tokens.match("(") )
         throw new ParseException("expected '(': " + tokens);

      if ( ! tokens.match("var") )
         throw new ParseException("expected 'var': " + tokens);

      Tree result = new Tree("var");

      // parse the variable
      if ( tokens.hasToken() )
         result.addSubTree( new Tree(tokens.nextToken()) );
      else
         throw new ParseException("expected variable name: " + tokens);

      // parse the expression
      if ( tokens.hasToken() )
         result.addSubTree( getExp(tokens) );
      else
         throw new ParseException("expected variable value: " + tokens);

      if ( (! tokens.hasToken()) || (! tokens.match(")")) )
         throw new ParseException("expected ')': " + tokens);

      return result;
   }//getVar()


   // Parse a begin expression
   private static Tree getBegin(Tokenizer tokens) throws ParseException
   {
      if ( ! tokens.match("(") )
         throw new ParseException("expected '(': " + tokens);

      if ( ! tokens.match("begin") )
         throw new ParseException("expected 'begin': " + tokens);

      Tree result = new Tree("begin");

      while ( tokens.hasToken() && (! tokens.peekToken().equals(")")) )
      {
         result.addSubTree( getExp(tokens) );
      }

      if ( (! tokens.hasToken()) || (! tokens.match(")")) )
         throw new ParseException("expected ')': " + tokens);

      return result;
   }//getBegin()


   // Parse a print expression
   private static Tree getPrint(Tokenizer tokens) throws ParseException
   {
      if ( ! tokens.match("(") )
         throw new ParseException("expected '(': " + tokens);

      if ( ! tokens.match("print") )
         throw new ParseException("expected 'print': " + tokens);

      Tree result = new Tree("print");

      // parse the expression
      if ( tokens.hasToken() )
         result.addSubTree( getExp(tokens) );
      else
         throw new ParseException("expected print's expression: " + tokens);

      if ( (! tokens.hasToken()) || (! tokens.match(")")) )
         throw new ParseException("expected ')': " + tokens);

      return result;
   }//getPrint()


   // Parse a boolean expression
   private static Tree getBexp(Tokenizer tokens) throws ParseException
   {
      if ( ! tokens.match("(") )
         throw new ParseException("expected '(': " + tokens);

      String tk = tokens.nextToken();

      Tree result = new Tree(tk);

      if ( tk.equals("&&") ||  tk.equals("||") )
      {
         // parse the first expression
         if ( tokens.hasToken() )
            result.addSubTree( getExp(tokens) );
         else
            throw new ParseException("expected "+tk+"'s first operand: " + tokens);

         // parse the second expression
         if ( tokens.hasToken() )
            result.addSubTree( getExp(tokens) );
         else
            throw new ParseException("expected "+tk+"'s second operand: " + tokens);

         // parse any other expressions
         while ( tokens.hasToken() && (! tokens.peekToken().equals(")")) )
         {
            result.addSubTree( getExp(tokens) );
         }
      }
      else if ( tk.equals("!") )
      {
         // parse the expression
         if ( tokens.hasToken() )
            result.addSubTree( getExp(tokens) );
         else
            throw new ParseException("expected !'s operand: " + tokens);
      }
      else // shouldn't get here
         throw new ParseException("unexpected boolean operator: " + tk);

      if ( (! tokens.hasToken()) || (! tokens.match(")")) )
         throw new ParseException("expected ')': " + tokens);

      return result;
   }//getBexp()


   // Parse a relational expression (which is a kind of boolean expression)
   private static Tree getRexp(Tokenizer tokens) throws ParseException
   {
      if ( ! tokens.match("(") )
         throw new ParseException("expected '(': " + tokens);

      String tk = tokens.nextToken();

      Tree result = new Tree(tk);

      // parse the first expression
      if ( tokens.hasToken() )
         result.addSubTree( getExp(tokens) );
      else
         throw new ParseException("expected "+tk+"'s first operand: " + tokens);

      // parse the second expression
      if ( tokens.hasToken() )
         result.addSubTree( getExp(tokens) );
      else
         throw new ParseException("expected "+tk+"'s second operand: " + tokens);

      if ( (! tokens.hasToken()) || (! tokens.match(")")) )
         throw new ParseException("expected ')': " + tokens);

      return result;
   }//getRexp()


   // Parse an arithmetic expression
   private static Tree getAexp(Tokenizer tokens) throws ParseException
   {
      if ( ! tokens.match("(") )
         throw new ParseException("expected '(': " + tokens);

      String tk = tokens.nextToken();

      Tree result = new Tree(tk);

      if ( tk.equals("+") )
      {
         // parse the first expression
         if ( tokens.hasToken() )
            result.addSubTree( getExp(tokens) );
         else
            throw new ParseException("expected "+tk+"'s first operand: " + tokens);

         // parse any other expression            s
         while ( tokens.hasToken() && (! tokens.peekToken().equals(")")) )
         {
            result.addSubTree( getExp(tokens) );
         }
      }
      else if ( tk.equals("-") )
      {
         // parse the first expression
         if ( tokens.hasToken() )
            result.addSubTree( getExp(tokens) );
         else
            throw new ParseException("expected "+tk+"'s first operand: " + tokens);

         // parse a second expression
         if ( tokens.hasToken() && (! tokens.peekToken().equals(")")) )
            result.addSubTree( getExp(tokens) );
      }
      else if ( tk.equals("*") )
      {
         // parse the first expression
         if ( tokens.hasToken() )
            result.addSubTree( getExp(tokens) );
         else
            throw new ParseException("expected "+tk+"'s first operand: " + tokens);

         // parse the second expression
         if ( tokens.hasToken() )
            result.addSubTree( getExp(tokens) );
         else
            throw new ParseException("expected "+tk+"'s second operand: " + tokens);

         // parse any other expression            s
         while ( tokens.hasToken() && (! tokens.peekToken().equals(")")) )
         {
            result.addSubTree( getExp(tokens) );
         }
      }
      else if ( tk.equals("/") || tk.equals("%") || tk.equals("^") )
      {
         // parse the first expression
         if ( tokens.hasToken() )
            result.addSubTree( getExp(tokens) );
         else
            throw new ParseException("expected "+tk+"'s first operand: " + tokens);

         // parse the second expression
         if ( tokens.hasToken() )
            result.addSubTree( getExp(tokens) );
         else
            throw new ParseException("expected "+tk+"'s second operand: " + tokens);
      }
      else // shouldn't get here
         throw new ParseException("expected arithmetic operator: " + tk);

      if ( (! tokens.hasToken()) || (! tokens.match(")")) )
         throw new ParseException("expected ')': " + tokens);

      return result;
   }//getAexp()
}//Parser